// SortListBoxOD.cpp : implementation file
// Copyright (c) 2016 by David Wincelberg
//
// Window styles:
// Required: owner-draw variable
// Recommended: horizontal scroll, vertical scroll, disable no scroll
// Optional: sort
// Prohibited: owner-draw fixed, has strings (so sorting will call CompareItem)
//
// Includes horizontal scrolling (based on the last reference).
//
// References:
// 1) J. Prosise, "Programming Windows with MFC -- 2nd ed."
// 2) MSDN sample CTRLTEST
// https://msdn.microsoft.com/en-us/subscriptions/downloads/zz9355ha%28v=vs.80%29.aspx
// 3) "Graphic and text ownerdraw listbox" at
// http://www.codeproject.com/KB/combobox/MFC_OwnerDraw_Listbox.aspx
// 4) "How to implement horizontal scrolling in a CListBox-derived class" at
// https://support.microsoft.com/en-us/kb/146437
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SortListBoxOD.h"

#include "ListBoxODData.h"
#include "ListBoxODOrder.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSortListBoxOD

// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortListBoxOD::CSortListBoxOD (CListBoxODOrder* pListBoxODOrder /* = nullptr */,
								bool bHandleDelete /* = true */)
{
	m_pListBoxODOrder = (pListBoxODOrder == nullptr ? new CListBoxODOrder() : pListBoxODOrder);
	m_bHandleDelete   = bHandleDelete;
	m_bSorting        = false;
	m_nHeight         = 0;
	m_bAllowOrderMod  = true;

	// Stores pointer for horizontal-extent calculations.
	SetBoxPointer (this);
}	// CSortListBoxOD


// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortListBoxOD::CSortListBoxOD (CStringOrder* pStringOrder,
								bool bHandleDeleteInner /* = true */,
								bool bHandleDeleteOuter /* = true */)
{
	m_pListBoxODOrder = new CListBoxODOrder (pStringOrder, bHandleDeleteInner);
	m_bHandleDelete   = bHandleDeleteOuter;
	m_bSorting        = false;
	m_nHeight         = 0;
	m_bAllowOrderMod  = true;

	// Stores pointer for horizontal-extent calculations.
	SetBoxPointer (this);
}	// CSortListBoxOD


CSortListBoxOD::~CSortListBoxOD()
{
	if (m_bHandleDelete)
		delete m_pListBoxODOrder;
}	// ~CSortListBoxOD


BEGIN_MESSAGE_MAP(CSortListBoxOD, CListBox)
	//{{AFX_MSG_MAP(CSortListBoxOD)
	//}}AFX_MSG_MAP
	ON_MESSAGE (LB_SETTABSTOPS, OnSetTabStops)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSortListBoxOD message handlers

// Action:	Checks for required styles and sets the sorting flag.
void CSortListBoxOD::PreSubclassWindow()
{
	DWORD dwStyle = GetStyle();
	if ((dwStyle & LBS_OWNERDRAWVARIABLE) == 0 ||
	(dwStyle & (LBS_OWNERDRAWFIXED | LBS_HASSTRINGS))) {
		TRACE (_T("** For CSortListBoxOD, set style owner-draw variable.\n"));
		TRACE (_T("** Do not set styles owner-draw fixed or has strings.\n"));
		ASSERT (false);
	}

	m_bSorting = ((dwStyle & LBS_SORT) != 0);

	CListBox::PreSubclassWindow();
}	// PreSubclassWindow


// Action:	Replaces the sort-order object.
//			Deletes the old one unless m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortListBoxOD::SetSortOrder (CListBoxODOrder* pListBoxODOrder,
								   bool bHandleDelete /* = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the sort-order strategy is not allowed ")
			_T("after adding or inserting a string.\n"));
		ASSERT (false);
		return;
	}

	if (pListBoxODOrder == nullptr)
		return;

	if (m_bHandleDelete)
		delete m_pListBoxODOrder;

	m_pListBoxODOrder = pListBoxODOrder;
	m_bHandleDelete   = bHandleDelete;
}	// SetSortOrder


// Action:	Replaces the string-order object.
//			Deletes the old one unless CListBoxODOrder's m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortListBoxOD::SetStringOrder (CStringOrder* pStringOrder,
									 bool bHandleDelete /*  = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the string-order strategy is not allowed ")
			_T("after adding or inserting a string.\n"));
		ASSERT (false);
		return;
	}

	if (pStringOrder == nullptr || m_pListBoxODOrder == nullptr)
		return;

	m_pListBoxODOrder->SetStringOrder (pStringOrder, bHandleDelete);
}	// SetStringOrder


// Action:	Provides the height of items.
// Note:	Even though the heights are the same, must use
//			LBS_OWNERDRAWVARIABLE for this function to be called.
void CSortListBoxOD::MeasureItem (LPMEASUREITEMSTRUCT lpMeasureItemStruct)
{
	lpMeasureItemStruct->itemHeight = GetHeight();
}	// MeasureItem


// Action:	Returns the height of list-box string items.
// Note:	tmAscent matches the height of an item in a regular list-box.
int CSortListBoxOD::GetHeight (bool bRecalc /* = false */)
{
	if (m_nHeight == 0 || bRecalc) {
		CDC* pDC = GetDC();
		TEXTMETRIC tm;
		pDC->GetTextMetrics (&tm);
		m_nHeight = tm.tmAscent;
	}

	return m_nHeight;
}	// GetHeight


// Action:	Checks if a list-box-data object was supplied; o.w. allocates one.
bool CSortListBoxOD::CheckListBoxData (CListBoxODData** ppListBoxODData, LPCTSTR lpszItem)
{
	if (*ppListBoxODData == nullptr) {
		try {
			*ppListBoxODData = new CListBoxODData (lpszItem);
		}
		catch (CMemoryException* e) {
			e->Delete();
			return false;
		}
	}

	return true;
}	// CheckListBoxData


// Action:	Adds a string to a list box without LBS_HASSTRINGS set.
// Note:	The string will be in the item data.  Using SetItemData
//			will not result in the data being stored in time for
//			CompareItem calls.
int CSortListBoxOD::AddString (LPCTSTR lpszItem,
							   CListBoxODData* pListBoxODData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListBoxODData;
		return LB_ERR;
	}

	if (!CheckListBoxData (&pListBoxODData, lpszItem))
		return LB_ERR;

	// Stores the data item instead of the string.
	// The documentation for CListBox::AddString says that CompareItem
	// is called if a list box has LBS_SORT but not LBS_HASSTRINGS.
	int nResult = CListBox::AddString (reinterpret_cast<LPCTSTR> (pListBoxODData));
	if (nResult == LB_ERR || nResult == LB_ERRSPACE) {
		delete pListBoxODData;
		return nResult;
	}

	StoreExtent (-1, lpszItem, pListBoxODData);	// The data item is already stored.
	m_bAllowOrderMod = false;
	return nResult;
}	// AddString


// Action:	Inserts a string to a list box without LBS_HASSTRINGS set.
// Note:	The string will be in the item data.
int CSortListBoxOD::InsertString (int nIndex, LPCTSTR lpszItem,
								  CListBoxODData* pListBoxODData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListBoxODData;
		return LB_ERR;
	}

	if (!CheckListBoxData (&pListBoxODData, lpszItem))
		return LB_ERR;

	// Stores the data item instead of the string.
	int nResult = CListBox::InsertString (nIndex, reinterpret_cast<LPCTSTR> (pListBoxODData));
	if (nResult == LB_ERR || nResult == LB_ERRSPACE) {
		delete pListBoxODData;
		return nResult;
	}

	StoreExtent (-1, lpszItem, pListBoxODData);	// The data item is already stored.
	m_bAllowOrderMod = false;
	return nResult;
}	// InsertString


// Action:	Deletes a string and may adjust the horizontal extent and
//			may allow the sort and string orders to change.
int CSortListBoxOD::DeleteString (UINT nIndex)
{
	CListBoxODData* pListBoxODData = GetItemDataObjPtr (nIndex);
	int nResult;

	if (pListBoxODData != nullptr) {
		int nExtent = pListBoxODData->GetExtent();
		nResult = CListBox::DeleteString (nIndex);
		if (nResult == LB_ERR)
			return LB_ERR;

		bool bRecalc = (nExtent >= m_nMaxExtent);

		if (m_bDefer)
			m_bNeedRecalc |= bRecalc;
		else if (bRecalc)
			UpdateMaxExtent();
	}
	else
		nResult = CListBox::DeleteString (nIndex);

	if (GetCount() == 0)
		ResetContent();

	return nResult;
}	// DeleteString


// Action:	Retrieves the string from the data for the given item.
void CSortListBoxOD::GetText (int nIndex, CString &rString) const
{
	if (nIndex < 0 || nIndex >= GetCount())
		rString.Empty();
	else {
		CListBoxODData* pListBoxODData = GetItemDataObjPtr (nIndex);

		if (pListBoxODData == nullptr)
			rString.Empty();
		else
			pListBoxODData->GetString (rString);
	}
}	// GetText


// Action:	Retrieves the length of the string from the data for the given item.
int CSortListBoxOD::GetTextLen (int nIndex) const
{
	if (nIndex < 0 || nIndex >= GetCount())
		return LB_ERR;

	CString strItem;
	GetText (nIndex, strItem);

	return strItem.GetLength();
}	// GetTextLen


// Action:	Retrieves the string from the data for the given item.
// Note:	lpszBuffer must be large enough to hold the string.
int CSortListBoxOD::GetText (int nIndex, LPTSTR lpszBuffer) const
{
	if (lpszBuffer == nullptr || nIndex < 0 || nIndex >= GetCount())
		return LB_ERR;

	CString strItem;
	GetText (nIndex, strItem);

	::lstrcpy (lpszBuffer, strItem);

	return strItem.GetLength();
}	// GetText


// Action:	Draws the string along with the selection state and the focus rectangle.
void CSortListBoxOD::DrawItem (LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	CDC dc;
	dc.Attach (lpDrawItemStruct->hDC);
	CRect rect = lpDrawItemStruct->rcItem;
	int nIndex = lpDrawItemStruct->itemID;

	if (rect.bottom == 0)	// Fixes problem of an empty rectangle.
		rect.bottom = GetHeight();

	CRect rectText (rect);
	rectText.left += 2; 	// Shifts text over by two pixels.

	CBrush* pBrush = new CBrush;
	pBrush->CreateSolidBrush (::GetSysColor ((lpDrawItemStruct->itemState &
		ODS_SELECTED) ? COLOR_HIGHLIGHT : COLOR_WINDOW));
	dc.FillRect (rect, pBrush);
	delete pBrush;

	if (nIndex != static_cast<UINT> (-1)) {
		CString strItem;
		GetText (nIndex, strItem);

		if (lpDrawItemStruct->itemState & ODS_SELECTED) {
			COLORREF crOldTextColor = dc.GetTextColor();
			COLORREF crOldBkColor   = dc.GetBkColor();
			dc.SetTextColor (::GetSysColor (COLOR_HIGHLIGHTTEXT));
			dc.SetBkColor (::GetSysColor (COLOR_HIGHLIGHT));

			dc.DrawText (strItem, rectText, DT_SINGLELINE | DT_VCENTER);

			dc.SetTextColor (crOldTextColor);
			dc.SetBkColor (crOldBkColor);
		}
		else {
			dc.DrawText (strItem, rectText, DT_SINGLELINE | DT_VCENTER);
		}
	}

	if (lpDrawItemStruct->itemState & ODS_FOCUS)
		dc.DrawFocusRect (rect);

	dc.Detach ();
}	// DrawItem


// Action:	Compares two strings.
// Note:	Uses item data since indexes (or indices) might not be set.
//			The framework expects the result to be -1, 0 or 1.
int CSortListBoxOD::CompareItem (LPCOMPAREITEMSTRUCT lpCompareItemStruct)
{
	if (m_pListBoxODOrder == nullptr)
		return 0;

	CListBoxODData* pData1 = reinterpret_cast<CListBoxODData*>
		(lpCompareItemStruct->itemData1);

	CListBoxODData* pData2 = reinterpret_cast<CListBoxODData*>
		(lpCompareItemStruct->itemData2);

	return m_pListBoxODOrder->Compare (pData1, pData2);
}	// CompareItem


// Action:	Finds a string prefix.
int CSortListBoxOD::FindString (int nStartAfter, LPCTSTR lpszItem) const
{
	int nCount = GetCount();
	if (lpszItem == nullptr || nStartAfter < -1 || nStartAfter >= nCount)
		return LB_ERR;

	int nLength = ::lstrlen (lpszItem), n, nCompare;
	CString strItem;

	for  (n = nStartAfter + 1;  n < nCount;  ++n) {
		GetText (n, strItem);
		nCompare = ::_tcsncicmp (strItem, lpszItem, nLength);	// generic for _strnicmp
		if (nCompare == 0)
			return n;
	}

	for  (n = 0;  n <= nStartAfter;  ++n) {
		GetText (n, strItem);
		nCompare = ::_tcsncicmp (strItem, lpszItem, nLength);	// generic for _strnicmp
		if (nCompare == 0)
			return n;
	}

	return LB_ERR;
}	// FindString


// Action:	Finds a full string.
int CSortListBoxOD::FindStringExact (int nStartAfter, LPCTSTR lpszItem) const
{
	int nCount = GetCount();
	if (lpszItem == nullptr || nStartAfter < -1 || nStartAfter >= nCount)
		return LB_ERR;

	int n, nCompare;
	CString strItem;

	for  (n = nStartAfter + 1;  n < nCount;  ++n) {
		GetText (n, strItem);
		nCompare = ::_tcsicmp (strItem, lpszItem);	// generic for _stricmp
		if (nCompare == 0)
			return n;
	}

	for  (n = 0;  n <= nStartAfter;  ++n) {
		GetText (n, strItem);
		nCompare = ::_tcsicmp (strItem, lpszItem);	// generic for _stricmp
		if (nCompare == 0)
			return n;
	}

	return LB_ERR;
}	// FindStringExact


// Action:	Selects based on a string prefix.
// Note:	Only works for single-selection list boxes.
int CSortListBoxOD::SelectString (int nStartAfter, LPCTSTR lpszItem)
{
	DWORD dwStyle = GetStyle();
	if (dwStyle & (LBS_EXTENDEDSEL | LBS_MULTIPLESEL))
		return LB_ERR;

	int nFind = FindString (nStartAfter, lpszItem);
	if (nFind == LB_ERR)
		return LB_ERR;

	SetCurSel (nFind);
	return nFind;
}	// SelectString


// Action:	Releases memory that is no longer needed.
// Note:	The subclass destructor will be called.
void CSortListBoxOD::DeleteItem (LPDELETEITEMSTRUCT lpDeleteItemStruct)
{
	delete reinterpret_cast<CListBoxODData*> (lpDeleteItemStruct->itemData);
}	// DeleteItem


// Action:	Clears the list box, resets the horizontal extent
//			and allows for the sort and string orders to be changed.
void CSortListBoxOD::ResetContent()
{
	CListBox::ResetContent();
	SetHorizontalExtent (0);
	m_nMaxExtent     = 0;
	m_bDefer         = false;
	m_bNeedRecalc    = false;
	m_bAllowOrderMod = true;
}	// ResetContent


// Action:	Stores user data in a CListBoxODData object.
int CSortListBoxOD::SetItemData (int nIndex, DWORD dwItemData)
{
	CListBoxODData* pListBoxODData = GetItemDataObjPtr (nIndex);

	if (pListBoxODData != nullptr) {
		pListBoxODData->SetItemData (dwItemData);
		return TRUE;
	}
	else
		return LB_ERR;
}	// SetItemData


// Action:	Retrieves user data from a CListBoxODData object.
DWORD CSortListBoxOD::GetItemData (int nIndex) const
{
	CListBoxODData* pListBoxODData = GetItemDataObjPtr (nIndex);

	if (pListBoxODData != nullptr)
		return pListBoxODData->GetItemData();
	else
		return LB_ERR;
}	// GetItemData


// Action:	Stores user data in a CListBoxODData object.
int CSortListBoxOD::SetItemDataPtr (int nIndex, void* pData)
{
	CListBoxODData* pListBoxODData = GetItemDataObjPtr (nIndex);

	if (pListBoxODData != nullptr) {
		pListBoxODData->SetItemData (reinterpret_cast<DWORD>(pData));
		return TRUE;
	}
	else
		return LB_ERR;
}	// SetItemDataPtr


// Action:	Retrieves user data from a CListBoxODData object.
void* CSortListBoxOD::GetItemDataPtr (int nIndex) const
{
	CListBoxODData* pListBoxODData = GetItemDataObjPtr (nIndex);

	if (pListBoxODData != nullptr)
		return reinterpret_cast<void*> (pListBoxODData->GetItemData());
	else
		return reinterpret_cast<void*> (LB_ERR);
}	// GetItemDataPtr


// Action:	Stores user data as a pointer to a CListBoxODData object.
// Note:	Copies the data values from the old data object and deletes it.
//			pListBoxODData should be from a subclass of CListBoxODData.
int CSortListBoxOD::SetItemDataObjPtr (int nIndex, CListBoxODData* pListBoxODData)
{
	if (pListBoxODData == nullptr)
		return LB_ERR;

	CListBoxODData* pOldListBoxData = GetItemDataObjPtr (nIndex);

	if (pOldListBoxData != nullptr) {
		pListBoxODData = pOldListBoxData;
		delete pOldListBoxData;
		return CListBox::SetItemData (nIndex, reinterpret_cast<DWORD> (pListBoxODData));
	}
	else
		return LB_ERR;
}	// SetItemDataObjPtr


// Action:	Retrieves user data as a pointer to a CListBoxODData object.
CListBoxODData* CSortListBoxOD::GetItemDataObjPtr (int nIndex) const
{
	int nCount = GetCount();

	if (nCount == LB_ERR || nIndex < 0 || nIndex >= nCount)
		return nullptr;

	DWORD dwData = CListBox::GetItemData (nIndex);

	if (dwData != LB_ERR)
		return reinterpret_cast<CListBoxODData*> (dwData);
	else
		return nullptr;
}	// GetItemDataObjPtr

/////////////////////////////////////////////////////////////////////////////
// Extent functions

// Action:	Recalculates extents and item height if the font resources have changed.
// Note:	The top-level window receives the OnFontChange message.
//			It should pass it to listbox controls with extent calculations. 
void CSortListBoxOD::FontsChanged()
{
	RecalcHorizontalExtent();
	GetHeight (true);
}	// OnFontChange


// Action:	Intercepts the LB_SETTABSTOPS message to update the tab stop array.
LRESULT CSortListBoxOD::OnSetTabStops (WPARAM wParam, LPARAM lParam)
{
	LRESULT lResult = Default();
	if (!lResult)
		return lResult;

	return TabStopsSet (wParam, lParam, lResult);
}	// OnSetTabStops
